# coding: utf8
# Copyright (c) Météo France (2017-)
# This software is governed by the CeCILL-C license under French law.
# http://www.cecill.info
""" This module gathers CNRM specific functionalities. """

from __future__ import print_function
import numpy as np
import pandas as pd
import evaltools as evt
from datetime import datetime, date, timedelta, time
from os.path import isfile
import h5py
from functools import reduce


def updateFromDaimonobsHdf5(dataset, file_path, instr_name, correc_unit=1,
                            add_new_stations=True):
    """ Update nan values of a Dataset object with a Daimonobs hdf5.

    The path where to find the data in the hdf5 file must be
    <instr_name>/<dataset.species>/Y

    Parameters
    ----------
    dataset : evt.dataset.Dataset
        Dataset object on which the function is applied.
    file_path : str
        Path of the hdf5 file.
    instr_name : str
        Name of the first order group of the hdf5 file where
        to find the data.
    correc_unit : float
        Multiplicative factor applied to original values.
    add_new_stations : bool
        If True, new stations found in the input file are added
        to the Dataset.

    """
    import daimonobs

    instr = daimonobs.DefInstrument(
        instr_name,
        "SurfConcInstrument",
        species=[dataset.species])
    instr.Read(file_path)

    df = pd.DataFrame({'lat': instr.lats,
                       'lon': instr.lons,
                       'time': instr.time,
                       'values': instr.conc[dataset.species]*correc_unit})

    metadata = None
    if metadata is None:
        df['code'] = ["{0:.2f}".format(lat) + "/" + "{0:.2f}".format(lon)
                      for lat, lon in zip(df['lat'], df['lon'])]

    df = df.pivot(index='time', columns='code', values='values')
    if dataset.seriesType == 'hourly':
        df.index = df.index.floor('H')
    elif dataset.seriesType == 'daily':
        df.index = df.index.floor('D')

    if not df.index.is_unique:
        df = df.groupby(df.index).mean()

    if add_new_stations is True:
        dataset.addNewStations(df.columns)

    dataset.data.update(df)


def writeDaimonobsHdf5(dataset, file_path, instr_name):
    """ Write dataset in a hdf5 file compatible with daimonobs.

    Parameters
    ----------
    dataset : evt.dataset.Dataset
        Dataset object on which the function is applied.
    file_path : str
        Path of the output file.
    instr_name : str
        Name of the first order group of the hdf5 file where
        to store the data.

    """
    import daimonobs

    instr = daimonobs.DefInstrument(
        instr_name,
        "SurfConcInstrument",
        species=[dataset.species])

    df = dataset.data
    df = df.stack().to_frame()
    df.reset_index(level=0, inplace=True)
    df.columns = ['time', 'conc']
    df = df.merge(dataset.stations, how='left', left_index=True,
                  right_index=True)

    nobs = df.shape[0]
    instr.nobs += nobs  # increase the instrument counter
    if nobs > 0:
        instr.conc[dataset.species] = np.ma.concatenate(
            (instr.conc[dataset.species],
             np.ma.array(df.conc, dtype=instr.conc[dataset.species].dtype)))
        instr.lats = np.ma.concatenate(
            (instr.lats, np.ma.array(df.lat, dtype=instr.lats.dtype)))
        instr.lons = np.ma.concatenate(
            (instr.lons, np.ma.array(df.lon, dtype=instr.lons.dtype)))
        instr.time = np.ma.concatenate(
            (instr.time, np.ma.array(df.time, dtype=instr.np_typetime)))
        # if instr.hascov:
        #     instr.cov[dataset.species] = np.ma.concatenate(
        #         (instr.cov[dataset.species],
        #          np.ma.array(res['cov'],
        #                      dtype=instr.cov[dataset.species].dtype)))
        # By default all obs are accepted
        instr.mask = np.concatenate(
            (instr.mask, np.zeros((nobs), dtype=bool)))

    instr.Write(file_path)


def read_values_from_csv(dataset, generic_file_path, correc_unit=1,
                         usecols=[0, 1, 6], sep=';'):
    """ Collect data from csv files.

    Modify Dataset.data in place using non-NA values from
    csv files with no header containing at least 3 columns:
    station codes, times and concentration values.

    Parameters
    ----------
    dataset : evt.dataset.Dataset
        Dataset object on which the function is applied.
    generic_file_path : str
        Path of the files with {year}, {month}, {day} instead
        of the year, month and day number, and {species} instead
        instead of the species name.
    correc_unit : float
        Multiplicative factor applied to original values.
    usecols : list of 3 integer
        List of column indices corresponding to ['station code', 'time',
        'value']. Column numbering begins at 0.
    sep : str
        Column separator used in the csv.

    """
    if dataset.seriesType != 'hourly':
        raise evt.EvaltoolsError("Dataset series type must be " +
                                 "'hourly' to use this method !!!")

    for d in range(-1, dataset.nb_days):  # -1 since an EEA csv contains
        # 1, 2, ..., 23 , 0 (of D+1)
        currentDate = dataset.startDate+timedelta(days=d)
        f_path = generic_file_path.format(
            species=dataset.species,
            year=currentDate.year,
            month=currentDate.strftime('%m'),
            day=currentDate.strftime('%d'))
        if isfile(f_path):
            df = pd.read_csv(
                f_path,
                sep=';',
                names=['station', 'time', 'obs'],
                header=None,
                usecols=usecols)
            df.time = pd.to_datetime(df.time, format='%Y%m%d%H%M%S')
            df.drop_duplicates(inplace=True)
            try:
                df = df.pivot(index='time',
                              columns='station',
                              values='obs')
                if correc_unit != 1:
                    df = df*correc_unit
                dataset.data.update(df)
            except ValueError as err:
                print("*** Problem with file {file} (this file is "
                      "skipped): {err} ".format(file=f_path, err=err))
        else:
            print("*** Missing file: {}".format(f_path))


read_EEA_obs_csv = read_values_from_csv


class StoreH5(object):
    """ Tool designed for storing hourly time series data in HDF5 files.

    A file created with this class has a special architecture. It can
    hold data from different species on a same time period. The time
    period covered has a fixed length but can be shifted day by day.
    Boundary dates are saved in the file attributes.

    For each species, several dataset can be stored containing data
    for a same station list that can be increased.

    Parameters
    ----------
    file_path : str
        Path of the hdf5 file to read. The file must exist.

    """
    char_dtype = h5py.special_dtype(vlen=str)

    def __init__(self, file_path, read_only=False):
        # open file
        if read_only is True:
            self.file = h5py.File(file_path, 'r')
        else:
            self.file = h5py.File(file_path, 'r+')

        # load dates
        self.startDate = date(int(self.file.attrs['start'][0:4]),
                              int(self.file.attrs['start'][4:6]),
                              int(self.file.attrs['start'][6:8]))
        self.endDate = date(int(self.file.attrs['end'][0:4]),
                            int(self.file.attrs['end'][4:6]),
                            int(self.file.attrs['end'][6:8]))
        self.nb_days = int((self.endDate - self.startDate).days + 1)

        try:
            self.seriesType = self.file.attrs['seriesType']
        except KeyError:
            self.file.attrs['seriesType'] = 'hourly'
            self.seriesType = self.file.attrs['seriesType']

        # nTimes is a number of hours (seriesType == 'hourly') or days
        # (seriesType == 'daily')
        if self.seriesType == 'hourly':
            self.nTimes = self.nb_days*24
        elif self.seriesType == 'daily':
            self.nTimes = self.nb_days
        else:
            raise evt.EvaltoolsError(
                "seriesType argument must be either 'hourly' or 'daily' !!!")

    @classmethod
    def newFile(cls, file_path, startDate, endDate, seriesType='hourly'):
        """ Build a Store object by writing a new hdf5 file.

        If the file already exists, it is truncated.

        Parameters
        ----------
        file_path : str
            Path of the file to create.
        startDate : datetime.date
            Starting date of the file data.
        endDate : datetime.date
            Ending date of the file data.

        Returns
        -------
            New Store object.

        """
        new_file = h5py.File(file_path, 'w')
        new_file.attrs['start'] = startDate.strftime('%Y%m%d')
        new_file.attrs['end'] = endDate.strftime('%Y%m%d')
        new_file.attrs['seriesType'] = seriesType
        return cls(file_path)

    def summary(self):
        """ Print summary statistics on the hdf5 file data.

        Nan rate, minimum and maximum are computed for each station. Then,
        the minimum, maximum and median of each of these statistics are
        displayed.

        """
        print("File = {}".format(self.file))
        print("Period = {startDate} - {endDate}".format(
            startDate=self.startDate, endDate=self.endDate))
        print("Time step = {}".format(self.seriesType))
        for species in self.file.keys():
            print("")
            print("---------- {} ----------".format(species))
            print("number of stations = {}".format(
                self.file[species]['stations']['code'].shape[0]))
            for model in self.file[species]['data'].keys():
                print("")
                print("* {} *".format(model))
                ds = self.getDataset(species=species,
                                     dsetName=model,
                                     startDate=self.startDate,
                                     endDate=self.endDate)
                ds.summary()

    def newSpecies(self, species, stations, dsetNames=[]):
        """ Add a new species to the file of the Store object.

        Parameters
        ----------
        species : str
            Species name that will be the name of the new hdf5 file group.
        stations : list of str
            List of the station codes corresponding to the columns of the
            datasets to be stored.
        dsetNames : None or list of str
            List of the dataset names to initialize. The new dataset will be
            filled with np.nan.

        """
        grp = self.file.create_group(species)
        grp.create_group('stations')
        grp['stations'].create_dataset('code',
                                       data=np.array(stations,
                                                     dtype=self.char_dtype),
                                       maxshape=(None,))
        grp.create_group('data')
        for name in dsetNames:
            self.newDataset(species, name)

    def newDataset(self, species, dsetName):
        """ Add a new dataset to the file of the Store object.

        Parameters
        ----------
        species : str
            Species name that refered to the file group name
            where to add the dataset.
        dsetName : str
            Name of the new dataset. The new dataset will be
            filled with np.nan.

        """
        if species not in self.file.keys():
            raise evt.EvaltoolsError(
                ("{} is not a species of the hdf5 file. Please use " +
                 " newSpecies method to add it.").format(species))

        grp = self.file[species]
        grp['data'].create_dataset(
            dsetName,
            dtype='float32',
            data=np.full((self.nTimes, grp['stations']['code'].shape[0]),
                         np.nan),
            compression="gzip",
            compression_opts=9,
            maxshape=(self.nTimes, None),
            fillvalue=np.nan)

    def addStations(self, species, new_stations):
        """ Add stations to the species dataframes.

        Add columns filled with np.nan to all the dataset
        of the species.
        New station codes must not already be present.

        Parameters
        ----------
        species : str
            Species name that refered to the file group name
            where to add the new stations.
        new_stations : list of str
            List of the codes of the new stations.

        """
        if np.in1d(new_stations,
                   self.file[species]['stations']['code'][:]).any():
            raise evt.EvaltoolsError(
                "Trying to add already existing stations !!!")

        stations = np.concatenate((self.file[species]['stations']['code'][:],
                                   new_stations))
        nsta = stations.shape[0]
        self.file[species]['stations']['code'].resize((nsta,))
        self.file[species]['stations']['code'][
            len(stations)-len(new_stations):] = np.array(new_stations).astype(
                '|S7')

        for dset in self.file[species]['data'].values():
            dset.resize((self.nTimes, nsta))

    def update(self, dataset, dsetName, add_new_stations=True):
        """ Update a dataset of the hdf5 file with Dataset object.

        Modify a dataset of the hdf5 file using non-NA values from passed
        Dataset object. The species of the modified dataset is the
        species of the Dataset object.

        Parameters
        ----------
        dataset : Dataset object
            Dataset object containing new values to add to the file.
        dsetName : str
            Name of the dataset to update.
        add_new_stations : bool
            If true, stations only present in dataset are also
            added to the hdf5 file.

        """
        # check if series types match
        if self.seriesType != dataset.seriesType:
            raise evt.EvaltoolsError(
                "Series types of dataset and self must be equal !!!")

        # check if h5 file holds the appropriate dataset
        species = dataset.species
        if species not in self.file.keys():
            raise evt.EvaltoolsError(
                ("{} is not a species of the hdf5 file. Please " +
                 "use newSpecies method to add it.").format(species))

        if dsetName not in self.file[species]['data'].keys():
            raise evt.EvaltoolsError(
                ("{dset} is not a dataset name of {species} " +
                 "group of the hdf5 file. Please use newDataset method to " +
                 "add it.").format(dset=dsetName, species=species))

        # find indices of data to load from the hdf5 file
        if self.startDate > dataset.startDate:
            first_idx = 0
            startTime = datetime.combine(self.startDate, time(0))
        else:
            first_idx = (dataset.startDate - self.startDate).days*24
            startTime = datetime.combine(dataset.startDate, time(0))

        if self.endDate < dataset.endDate:
            last_idx = self.nTimes
        else:
            last_idx = self.nTimes - (self.endDate - dataset.endDate).days*24

        # get station list from h5 file and add some if missing
        stations = self.file[species]['stations']['code'][:]
        if add_new_stations:
            new_stations = dataset.data.columns[
                np.logical_not(np.in1d(dataset.data.columns, stations))]
            self.addStations(species, new_stations)
            stations = self.file[species]['stations']['code'][:]

        # construct datetime index
        if self.seriesType == 'hourly':
            times = pd.DatetimeIndex([startTime+timedelta(hours=h)
                                      for h in range(last_idx-first_idx)])
        elif self.seriesType == 'daily':
            times = pd.DatetimeIndex([startTime.date()+timedelta(days=d)
                                      for d in range(last_idx-first_idx)])

        # get h5 file dataset
        df = pd.DataFrame(self.file[species]['data'][dsetName][
                              first_idx:last_idx, :],
                          index=times,
                          columns=stations)

        # write updated dataset in the h5 file
        df.update(dataset.data)
        self.file[species]['data'][dsetName][
            first_idx:last_idx, :] = df.values

    def shiftOneDay(self):
        """ Shift the hdf5 file time period one day.

        In all datasets of the file, rows corresponding to the
        earliest day are removed and new rows containing np.nan
        are addes for the new day.

        """
        for spe in self.file.keys():
            grp = self.file[spe]['data']
            for model in grp.keys():
                arr = grp[model][24:(24*self.nb_days)]
                grp[model][0:(24*(self.nb_days-1))] = arr
                grp[model][(24*(self.nb_days-1)):(24*self.nb_days)] = np.nan
        self.startDate = self.startDate + timedelta(days=1)
        self.endDate = self.endDate + timedelta(days=1)
        self.file.attrs['start'] = self.startDate.strftime('%Y%m%d')
        self.file.attrs['end'] = self.endDate.strftime('%Y%m%d')

    def getDataset(self, species, dsetName, startDate, endDate):
        """ Get a dataset of the hdf5 file.

        Parameters
        ----------
        species : str
            Species name (corresponding to the first wanted key of
            the hdf5 file).
        dsetName : str
            Name of the dataset to get.
        startDate : datetime.date object
            The date from which data is collected.
        endDate : datetime.date object
            The data until which data is collected.

        Returns
        -------
            evaltools.dataset.Dataset

        """
        # find indices of data to load from the hdf5 file
        if self.startDate > startDate:
            first_idx = 0
            startTime = datetime.combine(self.startDate, time(0))
        else:
            first_idx = (startDate - self.startDate).days*(
                24*(self.seriesType == 'hourly') +
                1*(self.seriesType == 'daily'))
            startTime = datetime.combine(startDate, time(0))
        if self.endDate < endDate:
            last_idx = self.nTimes
        else:
            last_idx = self.nTimes - (self.endDate - endDate).days*(
                24*(self.seriesType == 'hourly') +
                1*(self.seriesType == 'daily'))

        # construct datetime index
        if self.seriesType == 'hourly':
            times = pd.DatetimeIndex([startTime+timedelta(hours=h)
                                      for h in range(last_idx-first_idx)])
        elif self.seriesType == 'daily':
            times = pd.DatetimeIndex([startTime.date()+timedelta(days=d)
                                      for d in range(last_idx-first_idx)])

        # get DataFrame
        df = pd.DataFrame(self.file[species]['data'][dsetName][
                              first_idx:last_idx, :],
                          index=times,
                          columns=self.file[species]['stations']['code'])

        # construct Dataset
        res = evt.dataset.Dataset(
            stations=df.columns,
            startDate=startDate,
            endDate=endDate,
            species=species,
            seriesType=self.seriesType,
        )
        res.data.update(df)

        return res


def obsFromHDF(startDate, endDate, species, hdf_path, listing_path,
               forecastHorizon=1, correc_unit=1, **kwargs):
    """ Construct an evaluator.Observations object from HDF5 files.

    Parameters
    ----------
    startDate : datetime.date
        Starting day of the studied period.
    endDate : datetime.date
        Ending day (included) of the studied period.
    species : str
        Species name (must be a key of the hdf file).
    hdf_path : str
        Path of the input file. Input data can be splitted over
        two hdf files corresponding to two different years. In
        that case, the file path must contain '{year}'.
    listing_path : str
        Path of the station listing. This listing is only used
        to get metadata of the stations.
    forecastHorizon : int
        Number of forecasted days.
    correc_unit : float
        Factor to apply to original values.
    **kwargs :
        These parameters (like 'sep', 'sub_list', ...) will be
        passed to evaltools.utils.read_listing().

    Returns
    -------
        evaltools.evaluator.Observations object

    """
    store_list = [evt.dataframe.Store(hdf_path.format(year=year),
                                      read_only=True)
                  for year in np.unique([startDate.year, endDate.year])]
    seriesType = store_list[0].seriesType
    ds_list = [store.getDF(
        species=species,
        dsetName='observations',
        startDate=startDate,
        endDate=endDate+timedelta(days=forecastHorizon-1))
               for store in store_list]
    obs_ds = evt.dataset.Dataset(
        stations=reduce(np.union1d, [d.DF.columns for d in ds_list]),
        startDate=startDate,
        endDate=endDate+timedelta(days=forecastHorizon-1),
        species=species,
        seriesType=seriesType)
    for ds in ds_list:
        obs_ds.updateFromDataset(ds)
    observations = evt.evaluator.Observations.fromDataset(
        obs_ds, listing_path=listing_path,
        forecastHorizon=forecastHorizon, correc_unit=correc_unit,
        **kwargs)

    return observations


def simFromHDF(startDate, endDate, species, model, stations, hdf_path,
               stationsIdx, forecastHorizon=1, correc_unit=1):
    """ Construct an evaluator.Simulations object from HDF5 files.

    Parameters
    ----------
    startDate : datetime.date
        Starting day of the studied period.
    endDate : datetime.date
        Ending day (included) of the studied period.
    species : str
        Species name (must be a key of the hdf file).
    model : str
        Name of the model.
    stations : 1D array of str
        List of stations to keep in the resulting object.
    hdf_path : str
        Path of the input file. The file path must contain
        '{forecastDay}' instead of the forecast day of the file.
        Input data can be splitted over two hdf files
        corresponding to two different years. In that case, the
        file path must contain '{year}'.
    stationsIdx : list of str
        List of the names of studied stations.
    forecastHorizon : int
        Number of forecasted days.
    correc_unit : float
        Factor to apply to original values.

    Returns
    -------
        evaltools.evaluator.Simulations object

    """
    sim_ds_list = []
    for fd in range(forecastHorizon):
        store_list = [evt.dataframe.Store(hdf_path.format(year=y,
                                                          forecastDay=fd),
                                          read_only=True)
                      for y in np.unique([startDate.year, endDate.year])]
        ds_list = [store.getDF(species=species,
                               dsetName=model,
                               startDate=startDate+timedelta(days=fd),
                               endDate=endDate+timedelta(days=fd))
                   for store in store_list]
        ds = evt.dataset.Dataset(
            stations=np.intersect1d(reduce(np.union1d,
                                           [ds.data.columns
                                            for ds in ds_list]),
                                    stations),
            startDate=startDate+timedelta(days=fd),
            endDate=endDate+timedelta(days=fd),
            species=species)
        for temp_ds in ds_list:
            ds.updateFromDataset(temp_ds)
        sim_ds_list.append(ds)

    simulations = evt.evaluator.Simulations.fromDataset(
        model=model,
        ds_list=sim_ds_list,
        stationsIdx=stationsIdx,
        correc_unit=correc_unit)

    return simulations


def evaluatorFromHDF(cls, startDate, endDate, species, model, stations,
                     obs_hdf_path, sim_hdf_path, listing_path,
                     forecastHorizon=1, correc_unit=1, color='k', **kwargs):
    """ Construct an evaluator.Evaluator object from HDF5 files.

    The architecture of HDF5 files must be the one used by
    :class:`~evaltools.cnrm.Store`.

    Parameters
    ----------
    startDate : datetime.date
        Starting day of the studied period.
    endDate : datetime.date
        Ending day (included) of the studied period.
    species : str
        Species name (must be a key of the hdf file).
    model : str
        Name of the model.
    stations : 1D array of str
        List of stations to keep in the resulting object.
    obs_hdf_path : str
        Path of the input file. Input data can be splitted over
        two hdf files corresponding to two different years. In
        that case, the file path must contain '{year}'.
    sim_hdf_path : str
        Path of the input file. The file path must contain
        '{forecastDay}' instead of the forecast day of the file.
        Input data can be splitted over two hdf files
        corresponding to two different years. In that case, the
        file path must contain '{year}'.
    listing_path : str
        Path of the station listing. This listing is only used
        to get metadata of the stations.
    forecastHorizon : int
        Number of forecasted days.
    correc_unit : float
        Factor to apply to original values.
    color : str
        Default color that will be used in plotting functions.
    **kwargs :
        These parameters (like 'sep', 'sub_list', ...) will be
        passed to evaltools.utils.read_listing().

    Returns
    -------
        evaltools.evaluator.Evaluator object

    """
    observations = evt.Observations.fromHDF(
        startDate, endDate, species, obs_hdf_path, listing_path,
        forecastHorizon=forecastHorizon, correc_unit=correc_unit,
        **kwargs)
    simulations = evt.Simulations.fromHDF(
        startDate, endDate, species, model, stations, sim_hdf_path,
        listing_path, forecastHorizon=forecastHorizon,
        correc_unit=correc_unit, **kwargs)
    return evt.evaluator.Evaluator(observations, simulations, color=color)
